from dataclasses import dataclass

from voir.tools import configurable, gated, instrument_definition, parametrized

from .common import program


@gated("--wow")
def wow(ov):
    yield ov.phases.run_script
    print("WOW!")


def test_hello_flags_on(ov, outlines):
    ov.require(wow)
    ov(["--wow", program("hello")])
    assert outlines() == ["hello world", "WOW!"]


def test_hello_flags_off(ov, outlines):
    ov.require(wow)
    ov([program("hello")])
    assert outlines() == ["hello world"]


@gated("--wow", "Turn on the WOW")
def wow2(ov):
    yield ov.phases.run_script
    print("WOW!")


def test_gated_with_doc(ov, outlines):
    ov.require(wow2)
    ov(["--wow", program("hello")])
    assert outlines() == ["hello world", "WOW!"]


@parametrized("--funk", type=int, help="How much funk?")
def funk(ov):
    yield ov.phases.run_script
    for i in range(ov.options.funk):
        print("F U N K!")


def test_parametrized(ov, outlines):
    ov.require(funk)
    ov(["--funk", "3", program("hello")])
    assert outlines() == [
        "hello world",
        "F U N K!",
        "F U N K!",
        "F U N K!",
    ]


@instrument_definition
def instrument1(ov, x, y, z):
    yield ov.phases.init
    ov.log({"x": x, "y": y, "z": z})


def test_instrument_definition(ov, check_all):
    ov.require(instrument1(5, 6, 7))
    ov([program("hello")])


@dataclass
class Configuration:
    x: bool = True
    zazz: int = 4


@configurable
def instrument2(ov, cfg: Configuration):
    yield ov.phases.init
    ov.log({"x": cfg.x, "zazz": cfg.zazz})


def test_configurable(ov, check_all):
    ov.require(instrument2)
    ov([program("hello")])


def test_configurable2(ov, check_all):
    ov.require(instrument2)
    ov(["--no-x", "--zazz", "89", program("hello")])


def test_configurable3(ov, progdir, check_all):
    ov.require(instrument2)
    ov(["--config", str(progdir / "cfg.yaml"), program("hello")])


def test_configurable4(ov, progdir, check_all):
    ov.require(instrument2)
    ov(["--config", str(progdir / "cfg.yaml"), "-x", program("hello")])



from giving import given

from voir.helpers import iterate
from voir.overseer import Overseer

from .common import program


def print_given(ov):
    yield ov.phases.init
    ov.given.print()


def test_iterate_program(check_all):
    ov = Overseer(instruments=[print_given])
    ov([program("iterate"), "0"])


def test_iterate_program_report_batch(check_all):
    ov = Overseer(instruments=[print_given])
    ov([program("iterate"), "1"])


def test_log(check_all, data_fds):
    _, w = data_fds
    ov = Overseer(instruments=[], logfile=w)
    ov([program("log")])


def _extractor(key):
    def xtract():
        return (yield)[key]

    return xtract


def test_iterate_batch_size():
    with given() as gv:
        bs = gv.wmap("step", _extractor("batch_size")).accum()
        gv.display()
        for _ in iterate("x", range(10), report_batch=True, batch_size=3):
            pass

    assert bs == [3] * 10


def test_iterate_batch_size_fn():
    with given() as gv:
        bs = gv.wmap("step", _extractor("batch_size")).accum()
        gv.display()
        for _ in iterate("x", range(10), report_batch=True, batch_size=lambda x: x + 1):
            pass

    assert bs == list(range(1, 11))


def test_iterate_no_length():
    def generate():
        for x in range(10):
            yield x

    with given() as gv:
        bs = gv.wmap("step", _extractor("batch_size")).accum()
        gv.display()
        for _ in iterate(
            "x", generate(), report_batch=True, batch_size=lambda x: x + 1
        ):
            pass

    assert bs == list(range(1, 11)) + [None]



"""Manage the execution of the program."""

from ..phase import StopProgram
from ..tools import instrument_definition


@instrument_definition
def early_stop(ov, key, n, task=None, signal=StopProgram):
    """Stop the program early after a certain number of events.

    For example, given this code:

    .. code-block:: python

        for i in range(100):
            give(x=i)

    And this instrument in the voirfile:

    .. code-block:: python

        def instrument_xyz(ov):
            yield ov.phases.init
            ov.require(early_stop(key="x", n=20))

    ``voir script.py`` will stop after it has seen the "x" event 20 times, so it will
    only print up to the number 19 and then halt.

    Arguments:
        key: The key to watch for.
        n: The number of events with that key after which to stop.
        task: If not None, filter events where the ``task`` key has this value.
        signal: The exception to raise (defaults to :class:`voir.phase.StopProgram`)
    """
    called = False

    def _stop(value):
        # The stop signal may have the unfortunate effect of creating
        # another event, so this may get called twice.
        nonlocal called
        if not called:
            called = True
            if isinstance(signal, str):
                ov.log({"$event": signal})
            else:
                raise signal(value)

    yield ov.phases.init

    stream = ov.given
    if task is not None:
        stream = stream.where(task=task)
    stream = stream.where(key)
    stream.map_indexed(
        lambda _, idx: {"task": "early_stop", "progress": (idx + 1, n)}
    ).give()
    stream.skip(n) >> _stop



"""Log values automatically from the ``given`` stream."""

import fnmatch

from ..tools import instrument_definition


def _keep(patterns, context):
    """Produce a function that filters a data dictionary with the patterns.

    Returns ``False`` if the data does not match.
    """

    def operation(data):
        result = {}
        ok = False
        for k, v in data.items():
            if k in context or any(fnmatch.fnmatch(k, p) for p in context):
                result[k] = v
            if k in patterns or any(fnmatch.fnmatch(k, p) for p in patterns):
                result[k] = v
                ok = True
        return ok and result

    return operation


@instrument_definition
def log(ov, *patterns, context=[]):
    """Forward data from :attr:`~give.overseer.Overseer.given` into :attr:`~give.overseer.Overseer.log`.

    For each data dictionary in the ``given`` stream, all keys that match at least
    one of the ``patterns`` will be forwarded. If at least one key is forwarded, we also
    forward any keys that are in ``context``.

    For example, if you have this instrument:

    .. code-block:: python

        def instrument_xyz(ov):
            yield ov.phases.init
            ov.require(log("x*", "y", context=["z"]))

    It will affect statements in the main script like this:

    .. code-block:: python

        give(x=2)             # Forwards {"x": 2}
        give(x=2, task=123)   # Forwards {"x": 2, "task": 123}
        give(x=2, a=3)        # Forwards {"x": 2} but not a
        give(xylophone=1)     # Forwards {"xylophone": 1}
        give(a=3)             # Nothing is forwarded
        give(a=3, task=123)   # Nothing is forwarded

    .. note::
        If a patterns start with ``+``, that is equivalent to being in the context
        list, i.e. you can write ``log("x", "+y")`` instead of ``log("x", context="y")``

    Arguments:
        patterns: Patterns for the keys to forward.
        context: Extra keys to forward ONLY IF at least one key matches the patterns.
    """

    if not isinstance(context, (list, tuple)):
        context = [context]

    yield ov.phases.init

    more_context = {p[1:] for p in patterns if p.startswith("+")}
    context = {*more_context, *context}
    patterns = {p for p in patterns if not p.startswith("+")}

    ov.given.map(_keep(patterns, context)).filter(lambda x: x) >> ov.log



import os
import traceback
from contextlib import contextmanager

from .common import NotAvailable

IMPORT_ERROR = None
try:
    import pyhlml
except ImportError as err:
    IMPORT_ERROR = err


def is_installed():
    return IMPORT_ERROR is None


#
# pyhlml use the same enum as nvidia
#
NVML_ERROR_NOT_SUPPORTED = 3
NVML_TEMPERATURE_GPU = 0

NVSMI_UUID = 6
NVSMI_INDEX = 17
NVSMI_NAME = 4
NVSMI_MEMORY_TOTAL = 50
NVSMI_MEMORY_FREE = 51
NVSMI_MEMORY_USED = 52
NVSMI_TEMPERATURE_GPU = 130
NVSMI_UTILIZATION_GPU = 60
NVSMI_UTILIZATION_MEM = 61
NVSMI_POWER_DRAW = 141


def fix_num(n):
    try:
        return float(n)
    except ValueError:
        return -1


def tostr(data):
    if isinstance(data, bytes):
        return data.decode("utf-8")
    return str(data)


def handle_error(err):
    if err.value == NVML_ERROR_NOT_SUPPORTED:
        return "N/A"
    else:
        return err.__str__()


def safecall(call, *args):
    try:
        return call(*args)
    except pyhlml.HLMLError as err:
        return handle_error(err)


def make_gpu_infos(handles, selection):
    gpu_infos = {}

    for gid, handle in handles.items():
        try:
            if info := make_gpu_info(gid, handle, selection):
                gpu_infos[gid] = info
        except pyhlml.HLMLError:
            traceback.print_exc()

    return gpu_infos


def make_gpu_info(gid, handle, selection):
    uuid = tostr(safecall(pyhlml.hlmlDeviceGetUUID, handle))

    is_selected = (selection is None) or (
        selection and (str(gid) in selection or uuid in selection)
    )
    if not is_selected:
        return {}

    memInfo = pyhlml.hlmlDeviceGetMemoryInfo(handle)
    util = pyhlml.hlmlDeviceGetUtilizationRates(handle)

    return {
        "minor_number": tostr(safecall(pyhlml.hlmlDeviceGetMinorNumber, handle)),
        "device": gid,
        "product": tostr(safecall(pyhlml.hlmlDeviceGetName, handle)),
        "memory": {
            "used": memInfo.used / 1024 / 1024,
            "total": memInfo.total / 1024 / 1024,
        },
        "utilization": {
            "compute": util / 100.0,
            "memory": memInfo.used / memInfo.total,
        },
        "temperature": fix_num(
            safecall(pyhlml.hlmlDeviceGetTemperature, handle, NVML_TEMPERATURE_GPU)
        ),
        "power": fix_num(safecall(pyhlml.hlmlDeviceGetPowerUsage, handle)) / 1000.0,
        "selection_variable": "HABANA_VISIBLE_MODULES",
    }


@contextmanager
def hlmlinit():
    try:
        pyhlml.hlmlInit()
        yield
        pyhlml.hlmlShutdown()
    except pyhlml.hlml_error.HLMLError_AlreadyInitialized:
        yield
    except Exception:
        raise NotAvailable()


class DeviceSMI:
    def _setup(self):
        self.handles = {}

        if IMPORT_ERROR is not None:
            raise IMPORT_ERROR

        with hlmlinit():
            deviceCount = pyhlml.hlmlDeviceGetCount()
            for i in range(0, deviceCount):
                self.handles[i] = pyhlml.hlmlDeviceGetHandleByIndex(i)

    def __init__(self) -> None:
        self.hlsmi = None
        self._setup()

    @property
    def arch(self):
        return "hpu"

    @property
    def visible_devices(self):
        return os.environ.get("HABANA_VISIBLE_MODULES", None)

    def get_gpus_info(self, selection=None):
        with hlmlinit():
            return make_gpu_infos(self.handles, selection)

    def close(self):
        pass

    def system_info(self):
        try:
            cuda_driver_version = pyhlml.hlmlSystemGetCudaDriverVersion_v2()
            driver_version = pyhlml.hlmlSystemGetDriverVersion()
            entries = pyhlml.hlmlSystemGetHicVersion()
            nvml_version = pyhlml.hlmlSystemGetNVMLVersion()
            return {
                "CUDA_DRIVER": cuda_driver_version,
                "DRIVER_VERSION": driver_version,
                "HIC_VERSION": entries,
                "NVML_VERSION": nvml_version,
            }
        except Exception:
            import traceback

            traceback.print_exc()
            return {}



"""Monitor IO usage."""

import psutil


def io_monitor(perdisk=False):
    start = psutil.disk_io_counters(perdisk=perdisk)
    busy_time = getattr(start, "busy_time", 0)

    def monitor():
        iocounters = psutil.disk_io_counters(perdisk=perdisk)

        def diskinfo(diskio):
            return {
                "read_count": diskio.read_count - start.read_count,
                "write_count": diskio.write_count - start.write_count,
                "read_bytes": diskio.read_bytes - start.read_bytes,
                "read_time": diskio.read_time - start.read_time,
                "write_time": diskio.write_time - start.write_time,
                "busy_time": getattr(diskio, "busy_time", 0) - busy_time,
            }

        if perdisk:
            return {str(k): diskinfo(diskio) for k, diskio in iocounters.items()}

        return diskinfo(iocounters)

    return monitor



def instrument_header(ov):
    yield ov.phases.init
    print("~~~~~")


def greet(ov):
    yield ov.phases.init
    print("<heyoo>")
    yield ov.phases.run_script
    print("<bye>")


instrument_list = [{"greet": greet}]



"""This module defines helper functions that are meant to be used inside programs.

There is typically no reason to use these in a voirfile, but they may be useful in
a script run through Voir.
"""

from contextvars import ContextVar

from giving import give

current_overseer = ContextVar("current_overseer", default=None)


def log(**kwargs):
    """Log key/values directly to the Overseer.

    An :class:`Overseer` typically has two main streams, :attr:`~voir.overseer.Overseer.given`
    and :attr:`~voir.overseer.Overseer.log`. This sends data to the latter, which is
    meant to be exported out of the program, so the datatypes should be serializable.
    """
    ov = current_overseer.get()
    if ov is not None:
        ov.log(kwargs)


def iterate(
    task: str, iterable, report_batch=False, ignore_loading=False, batch_size=None
):
    """Stream events along an iterative process.

    ``iterate`` iterates over the iterable, and while it does so it generates a progress
    report that is accessible from :attr:`~give.overseer.Overseer.given`. For example,
    let us say you have a script and a voirfile defined as follows:

    **script.py:**

    .. code-block:: python

        for x in voir.iterate("count", range(10), report_batch=True):
            print(x)

    **voirfile.py:**

    .. code-block:: python

        def instrument_things(ov):
            yield ov.phases.init

            ov.given.where("progress", task="count").print()

        def enter_exit():
            print("<Enter loop body>")
            yield
            print("<Exit loop body>")

        ov.given.where(task="count").wrap("step", enter_exit)

    You would get the following output:

    .. code-block::

        {'task': 'count', 'progress': (0, 10)}
        <Enter loop body>
        0
        <Exit loop body>
        {'task': 'count', 'progress': (1, 10)}
        <Enter loop body>
        1
        <Exit loop body>

        ...

        {'task': 'count', 'progress': (9, 10)}
        <Enter loop body>
        9
        <Exit loop body>
        {'task': 'count', 'progress': (10, 10)}

    **Compatible instruments:**

    * :func:`voir.instruments.dash<voir.instruments.dash.dash>` displays a graphical progress
      bar from the data provided by ``iterate``.
    * :func:`voir.instruments.rate<voir.instruments.metric.rate>` calculates the rate at which data is
      processed by dividing the number of items per iteration (batch_size) by
      how much time each iteration takes.

    Arguments:
        task: The name of the task that this iteration is doing (e.g. ``"train"`` if
            you are iterating over a training set, or anything else that is representative
            of what you are doing). This is how the data streamed from different calls
            to ``iterate`` can be identified.
        iterable: An arbitrary iterable sequence.
        report_batch: If ``True``, in addition to progress, each time the iterator yields
            a batch, ``iterate`` will wrap its own ``yield`` with
            :meth:`give.wrap("step", batch)<giving.gvr.Giver.wrap>`. Instruments can
            intercept these and do something before and after every batch.
        ignore_loading: If ``True``, (and if ``report_batch`` is also ``True``)
            the step wrapper will ignore any work that is done
            within the iterator (that is to say it will call ``next()`` outside of the
            step).
        batch_size:
            * If ``None``, the ``batch`` will be given.
            * If a function, it will be called on ``batch`` and the result will be
              given as ``batch_size``.
            * If a number, it will be given as ``batch_size`` directly.
    """
    assert isinstance(task, str)
    try:
        n = len(iterable)
    except TypeError:
        n = None

    def prog(i):
        if n is not None:
            give(progress=(i, n))

    i = 0
    with give.inherit(task=task):
        prog(0)
        it = iter(iterable)
        while True:
            if i == n:
                break
            i += 1

            def get_batch():
                batch = next(it)
                if batch_size is None:
                    kwargs = {"batch": batch}
                elif callable(batch_size):
                    kwargs = {"batch_size": batch_size(batch)}
                else:
                    kwargs = {"batch_size": batch_size}
                return batch, kwargs

            try:
                if not report_batch:
                    batch, kwargs = get_batch()
                    yield batch
                elif ignore_loading:
                    batch, kwargs = get_batch()
                    with give.wrap("step", **kwargs):
                        yield batch
                else:
                    empty_kwargs = (
                        {"batch": None} if batch_size is None else {"batch_size": None}
                    )
                    with give.wrap("step", **empty_kwargs) as extra:
                        batch, kwargs = get_batch()
                        extra.update(kwargs)
                        yield batch
            except StopIteration:
                break
            prog(i)



import time

from voir import give, iterate

if __name__ == "__main__":
    batches = [[i] * 10 for i in range(10)]
    for batch in iterate("stuff", batches, report_batch=True):
        time.sleep(0.1)
        give(loss=1 / (1 + sum(batch)))



def header(ov):
    yield ov.phases.init
    print("/////")


def greet(ov):
    yield ov.phases.init
    print("<bonjour>")
    yield ov.phases.run_script
    print("<bonsoir>")


__instruments__ = [
    {
        "greet": greet,
        "header": header,
    }
]



"""This module defines an extension to ArgumentParser.

:class:`ExtendedArgumentParser` can be used to add groups of options defined in
a dataclass, and also provide defaults to them that may come from configuration
files and the like.
"""

import ast
import inspect
import tokenize
import warnings
from argparse import ArgumentParser
from dataclasses import MISSING, dataclass, fields, is_dataclass
from functools import partial
from textwrap import dedent
from typing import Union

from omegaconf import OmegaConf
from ovld import meta, ovld


@dataclass
class Info:
    name: str
    type: type
    help: str
    annotation: object
    prefix: str
    parser: ArgumentParser
    group: ArgumentParser


def _dash(base):
    base = base.replace("_", "-")
    if len(base) == 1:
        return f"-{base}"
    else:
        return f"--{base}"


@ovld
def contribute(default: meta(is_dataclass), info: Info):
    """Populate info.parser based on the default given.

    If there is no default value for the argument, ``default`` takes the value
    ``MISSING``.

    The :class:`Info` object already contains the name of the field, as well as its
    type, help string, etc. Depending on the type of the argument, an argument or
    a subparser will be added to ``info.parser``.
    """
    # NOTE: This function is called using ``contribute[argtype, Info](...)`` which
    # dispatches to the relevant version of ``contribute`` thanks to ``ovld``.
    group = info.parser
    hlp = getattr(info.type, "__help__", None) or info.type.__name__
    if hlp:
        group = info.parser.add_argument_group(hlp)
    docs = get_attribute_docstrings(info.type)
    if info.prefix is None:
        pfx = ""
    else:
        pfx = f"{info.prefix}{info.name}."
    info.parser.constructors[pfx] = (info.type, f"{info.prefix or ''}{info.name}")
    for field in fields(info.type):
        contribute[field.type, Info](
            field.default if default is MISSING else getattr(default, field.name),
            Info(
                name=field.name,
                type=field.type,
                help=docs.get(field.name, None),
                annotation=None,
                prefix=pfx,
                parser=info.parser,
                group=group,
            ),
        )


@ovld
def contribute(default: bool, info: Info):  # noqa: F811
    hlp = info.help
    pth = f"{info.prefix}{info.name}"

    is_default = "(Default) " if default is True else ""
    info.group.add_argument(
        # f"--{pth}",
        _dash(pth),
        action="store_true",
        dest=pth,
        default=default,
        help=f"{is_default}{hlp}",
    )

    pth_no = f"{info.prefix}no-{info.name}"
    is_default = "(Default) " if default is False else ""
    info.group.add_argument(
        _dash(pth_no),
        action="store_false",
        dest=pth,
        default=default,
        help=hlp and f"{is_default}Do not {hlp[0].lower()}{hlp[1:]}",
    )


@ovld
def contribute(default: Union[int, float, str], info: Info):  # noqa: F811
    info.group.add_argument(
        _dash(f"{info.prefix}{info.name}"),
        type=info.type,
        default=None if default is MISSING else default,
        required=default is MISSING,
        metavar=info.name.upper(),
        help=info.help,
    )


def _expand(args, constructors):
    for name in sorted(constructors, key=len, reverse=True):
        ln = len(name)
        cons, dest = constructors[name]
        cons_args = {}
        field_names = {f.name for f in fields(cons)}
        for k, v in list(vars(args).items()):
            if k.startswith(name):
                field_name = k[ln:]
                if field_name in field_names:
                    cons_args[field_name] = v
                    delattr(args, k)
        obj = cons(**cons_args)
        setattr(args, dest, obj)
    return args


class ExtendedArgumentParser(ArgumentParser):
    """Argument parser that supports adding option groups as dataclasses.

    Works just like :class:`argparse.ArgumentParser`, plus the method
    :meth:`~ExtendedArgumentParser.add_from_model`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.constructors = {}
        self.base_configs = {}
        self.base_configs_locked = False
        self.used_base_configs = set()

    def merge_base_config(self, config: dict):
        """Merge default values for the configurations.

        This method must be called **before** ``add_from_model(...)``.

        Call this if e.g. you want to import the settings from a configuration file
        as if they had been given as command-line arguments. You can call this method
        multiple times if you have multiple sources of data. You do not need to set
        every option.

        Options provided on the command line will always override what
        is set through this method.

        (The reason ``merge_base_config`` must be called before ``add_from_model``
        is that it may modulate the command-line options, e.g. which ones are optional
        and which ones are required.)

        Arguments:
            config: A dictionary that contains default values for the options
                that will be contributed by ``add_from_model``.
        """
        if self.base_configs_locked:
            raise Exception(
                "Cannot merge a base config after add_from_model is called."
            )
        self.base_configs = OmegaConf.merge(self.base_configs, config)

    def add_from_model(self, dest: str, model: type, flatten: bool = True):
        """Add the attributes of the ``model`` dataclass as options.

        If, for example, we have this dataclass:

        .. code-block:: python

            @dataclass
            class Options:
                # Some number
                abc: int
                # A flag
                xyz: bool = False

        * ``add_from_model("opts", Options)`` will add ``--abc`` and ``--xyz``
          to the argument parser, using the comments as their documentation,
          and the values will be in ``args.opts.abc`` etc.
        * ``add_from_model("opts", Options, flatten=False)`` will add ``--opts.abc``
          and ``--opts.xyz`` (putting the values in the same place).
        * ``--abc`` will be required, ``--xyz`` will be optional because it has
          a default value.

        Arguments:
            dest: The destination key in the arguments Namespace returned by
                ``parse_args``.
            model: A dataclass.
            flatten: If True, the arguments declared in the dataclass will be
                available at the top level. Otherwise, they will have to be
                given as ``--dest-xyz``.
        """
        self.base_configs_locked = True

        if isinstance(model, type):
            typ = model
            model = MISSING
        else:
            typ = type(model)

        self.used_base_configs.add(dest)
        if dest in self.base_configs:
            assert model is MISSING
            model = OmegaConf.merge(OmegaConf.structured(typ), self.base_configs[dest])

        contribute[typ, Info](
            model,
            Info(
                name=dest,
                type=typ,
                help=None,
                annotation=None,
                prefix=None if flatten else dest,
                parser=self,
                group=self,
            ),
        )

    def _parse_known_args(self, *args, **kwargs):
        unused = set(self.base_configs) - self.used_base_configs
        if unused:
            warnings.warn(
                f"Configuration blocks {unused} were not used. Valid blocks are: {self.used_base_configs}. Did you forget a nesting level?"
            )
        ns, args = super()._parse_known_args(*args, **kwargs)
        _expand(ns, self.constructors)
        return ns, args


#############################
# Extracting the docstrings #
#############################


def scrape_comments(src: str):
    """Use the tokenizer module to fetch comments from source.

    Arguments:
        src: The source code.

    Returns:
        List of (line, column, "COMMENT", comment_string)
    """
    lines = bytes(src, encoding="utf8").splitlines(keepends=True)
    return [
        (*tok.start, "COMMENT", tok.string[1:].strip())
        for tok in tokenize.tokenize(partial(next, iter(lines)))
        if tok.type == tokenize.COMMENT
    ]


class AttributeVisitor(ast.NodeVisitor):
    """Walk an AST to gather variable and docstring tokens and their lines/columns."""

    def __init__(self):
        self.data = []
        self.prefix = None

    def add_data(self, node, kind, content):
        self.data.append((node.lineno, node.col_offset, kind, content))

    def visit_body(self, name, stmts):
        old_prefix = self.prefix
        if self.prefix is None:
            self.prefix = ""
        else:
            self.prefix += f"{name}."
        for stmt in stmts:
            if (
                isinstance(stmt, ast.Expr)
                and isinstance(stmt.value, ast.Constant)
                and isinstance(stmt.value.value, str)
            ):
                self.add_data(stmt, "DOC", stmt.value.value)
            else:
                self.visit(stmt)
        self.prefix = old_prefix

    def visit_ClassDef(self, node):
        if self.prefix is not None:
            self.add_data(node, "VARIABLE", f"{self.prefix}{node.name}")
        self.visit_body(node.name, node.body)

    def visit_FunctionDef(self, node):
        if self.prefix is not None:
            self.add_data(node, "VARIABLE", f"{self.prefix}{node.name}")
        self.visit_body(node.name, node.body)

    def visit_Assign(self, node):
        self.generic_visit(node, may_assign=True)

    def visit_AnnAssign(self, node):
        self.generic_visit(node, may_assign=True)

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store):
            self.add_data(node, "VARIABLE", f"{self.prefix}{node.id}")

    def generic_visit(self, node, may_assign=False):
        if isinstance(node, ast.stmt) and not may_assign:
            self.add_data(node, "OTHER", None)
        super().generic_visit(node)


def scrape_variables_and_docstrings(src: str):
    """Scrape the variables and docstrings from source code.

    Arguments:
        src: The source code.

    Returns:
        List of (line, column, "VARIABLE"|"DOC"|"OTHER", name_or_content)
    """
    visitor = AttributeVisitor()
    visitor.visit(ast.parse(src))
    return visitor.data


def get_attribute_docstrings(cls):
    """Get the docstrings for individual attributes of a class.

    Arguments:
        cls: The class for which we want to get attribute documentation.

    Returns:
        A dict from variable name to its associated docstring (after itself) and/or
        comment (above itself).
    """
    docs = {}
    current = None
    current_line = None
    for_next = []
    src = dedent(inspect.getsource(cls))
    # We concatenate comment tokens from the tokenizer with
    # variable/docstring tokens extracted using the ast module
    data = sorted(scrape_comments(src) + scrape_variables_and_docstrings(src))
    for line, _, kind, content in data:
        if kind == "COMMENT":
            if current is not None and current_line == line:
                docs[current].append(content)
            else:
                for_next.append(content)
        elif kind == "DOC" and current:
            docs[current].append(content)
        elif kind == "VARIABLE":
            docs[content] = for_next
            for_next = []
            current = content
            current_line = line
        elif kind == "OTHER":
            current = current_line = None
            for_next = []
    return {k: "\n".join(lines) for k, lines in docs.items()}



import os
import traceback

from .common import NotAvailable

IMPORT_ERROR = None
try:
    import pynvml
    from pynvml import NVMLError_DriverNotLoaded, NVMLError_LibraryNotFound
except ImportError as err:
    IMPORT_ERROR = err


def fix_num(n):
    try:
        return float(n)
    except ValueError:
        return -1


def tostr(data):
    if isinstance(data, bytes):
        return data.decode("utf-8")
    return str(data)


def handle_error(err):
    if err.value == pynvml.NVML_ERROR_NOT_SUPPORTED:
        return "N/A"
    else:
        return err.__str__()


def safecall(call, *args):
    try:
        return call(*args)
    except pynvml.NVMLError as err:
        return handle_error(err)


def make_gpu_info(gid, handle, selection):
    uuid = tostr(safecall(pynvml.nvmlDeviceGetUUID, handle))

    is_selected = (selection is None) or (
        selection and (str(gid) in selection or uuid in selection)
    )
    if not is_selected:
        return {}

    util = pynvml.nvmlDeviceGetUtilizationRates(handle)

    try:
        memInfo = pynvml.nvmlDeviceGetMemoryInfo(handle)
        memInfo = {
            "used": memInfo.used / 1024 / 1024,
            "total": memInfo.total / 1024 / 1024,
        }

    except pynvml.NVMLError_NotSupported:
        import psutil

        mem = psutil.virtual_memory()
        memInfo = {
            "used": mem.used / 1024 / 1024,
            "total": mem.total / 1024 / 1024,
        }

    return {
        "minor_number": tostr(safecall(pynvml.nvmlDeviceGetMinorNumber, handle)),
        "device": gid,
        "product": tostr(safecall(pynvml.nvmlDeviceGetName, handle)),
        "memory": memInfo,
        "utilization": {
            "compute": util.gpu / 100,
            "memory": util.memory,
        },
        "temperature": fix_num(
            safecall(
                pynvml.nvmlDeviceGetTemperature, handle, pynvml.NVML_TEMPERATURE_GPU
            )
        ),
        "power": fix_num(safecall(pynvml.nvmlDeviceGetPowerUsage, handle)) / 1000.0,
        "selection_variable": "CUDA_VISIBLE_DEVICES",
    }


def make_gpu_infos(handles, selection):
    gpu_infos = {}

    for gid, handle in handles.items():
        try:
            if info := make_gpu_info(gid, handle, selection):
                gpu_infos[gid] = info
        except Exception:
            traceback.print_exc()

    return gpu_infos


def is_installed():
    return IMPORT_ERROR is None


class DeviceSMI:
    def _setup(self):
        self.handles = {}

        if IMPORT_ERROR is not None:
            raise IMPORT_ERROR

        try:
            pynvml.nvmlInit()
        except NVMLError_LibraryNotFound as err:
            raise NotAvailable() from err

        except NVMLError_DriverNotLoaded as err:
            raise NotAvailable() from err

        deviceCount = pynvml.nvmlDeviceGetCount()
        for i in range(0, deviceCount):
            self.handles[i] = pynvml.nvmlDeviceGetHandleByIndex(i)

    def __init__(self) -> None:
        self._setup()

    @property
    def arch(self):
        return "cuda"

    @property
    def visible_devices(self):
        return os.environ.get("CUDA_VISIBLE_DEVICES", None)

    def get_gpus_info(self, selection=None):
        return make_gpu_infos(self.handles, selection)

    def system_info(self):
        cuda_driver_version = pynvml.nvmlSystemGetCudaDriverVersion_v2()
        driver_version = pynvml.nvmlSystemGetDriverVersion()
        entries = pynvml.nvmlSystemGetHicVersion()
        nvml_version = pynvml.nvmlSystemGetNVMLVersion()
        return {
            "CUDA_DRIVER": cuda_driver_version,
            "DRIVER_VERSION": driver_version,
            "HIC_VERSION": entries,
            "NVML_VERSION": nvml_version,
        }

    def close(self):
        pass



import os

import pytest

from voir.overseer import JsonlFileLogger

from .common import program


def _probe(ov):
    yield ov.phases.init
    ov.argparser.add_argument("--probe")
    yield ov.phases.load_script
    if ov.options.probe:
        ov.probe(ov.options.probe) >> ov.log


def _crash(ov):
    yield ov.phases.init
    ov.argparser.add_argument("--crash", action="store_true")
    yield ov.phases.load_script
    if ov.options.crash:
        raise ValueError("boom.")


def test_probe(ov, capsys, capdata):
    ov.require(_probe)
    ov(["--probe", "//main > greeting", program("hello")])
    assert capsys.readouterr().out == "hello world\n"
    assert '{"greeting": "hello"}' in capdata().split("\n")


def test_hello(ov, capsys, capdata, file_regression):
    ov([program("hello")])
    assert capsys.readouterr().out == "hello world\n"
    file_regression.check(capdata())


def test_collatz(ov, outlines):
    ov([program("collatz"), "-n", "13"])
    results = [int(x) for x in outlines()]
    assert results == [13, 40, 20, 10, 5, 16, 8, 4, 2]


def test_not_serializable(ov, outlines, capdata):
    ov.require(_probe)

    ov(["--probe", "//main > parser", program("collatz"), "-n", "13"])

    results = [int(x) for x in outlines()]
    assert results == [13, 40, 20, 10, 5, 16, 8, 4, 2]

    assert "$unserializable" in capdata()


def test_error_unknown_program(ov, output_summary, file_regression):
    unknown = program("unknown")
    with pytest.raises(FileNotFoundError):
        ov([unknown])

    file_regression.check(output_summary().replace(unknown, "X"))


def test_error_in_load(ov, check_all):
    with pytest.raises(ZeroDivisionError):
        ov([program("zero")])


def test_error_in_run(ov, check_all):
    with pytest.raises(ValueError):
        ov([program("collatz"), "-n", "blah"])


def test_overseer_crash(ov, check_all):
    ov.require(_crash)
    # Should not impede the program's execution
    ov(["--crash", program("collatz"), "-n", "13"])


def test_jsonl_logger_invalid_fd():
    _, w = os.pipe()
    with open(w, "w"):
        pass
    with pytest.raises(OSError):
        JsonlFileLogger(w)

    logger = JsonlFileLogger(w, require_writable=False)
    logger.log({"a": 1, "b": 2})


class Terrible:
    def __str__(self):
        raise Exception()

    __repr__ = __str__


def test_jsonl_logger_bad_str():
    r, w = os.pipe()
    logger = JsonlFileLogger(w)
    logger.log({"a": Terrible()})
    logger.close()
    assert open(r, "r").read() == '{"$unrepresentable": null}\n'



"""This module defines tools to easily make configurable instruments."""

import functools
import inspect
from functools import partial

from ovld import meta, ovld


@ovld
def gated(flag: str):  # noqa: F811
    """Decorate an instrument so that it is only activated when a ``--flag`` is given.

    .. code-block:: python

        @gated("--xyz", "Ex why zee.")
        def instrument_xyz(ov):
            ...

    Arguments:
        flag: The name of the flag, e.g. ``"--flag"``
        help: Description of the flag.
        instrument: The instrument to place behind the flag.

    **List of signatures:**
    """
    return partial(gated, flag)


@ovld
def gated(flag: str, help: str):  # noqa: F811
    return partial(gated, flag, help=help)


@ovld
def gated(flag: str, instrument: meta(callable), help: str = None):  # noqa: F811
    dest = flag

    def run(ov):
        yield ov.phases.init
        ov.argparser.add_argument(flag, action="store_true", dest=dest, help=help)
        yield ov.phases.parse_args
        if getattr(ov.options, dest):
            ov.require(instrument)

    run.instrument = instrument
    return run


@ovld
def parametrized(option: str, type=None, help=None, default=None):  # noqa: F811
    """Decorate an instrument to declare an ``--option``.

    .. code-block:: python

        @parametrized("--xyz", int, "Ex why zee.", 2)
        def instrument_xyz(ov):
            value = ov.options.xyz
            ...

    Arguments:
        option: The name of the option, e.g. ``"--option"``
        type: The type of the option.
        help: Description of the option.
        default: Default value for the option.
        instrument: The instrument to place behind the flag.

    **List of signatures:**
    """
    return partial(parametrized, option, type=type, help=help, default=default)


@ovld
def parametrized(  # noqa: F811
    option: str, instrument: meta(callable), type=None, help=None, default=None
):
    def run(ov):
        yield ov.phases.init
        ov.argparser.add_argument(option, type=type, help=help, default=default)
        yield ov.phases.parse_args
        ov.require(instrument)

    return run


def instrument_definition(fn):
    """Define a parametrizable instrument.

    .. note::
        Such an instrument is parameterized in code and not by command-line flags.

    .. code-block:: python

        @instrument_definition
        def wait_a_bit(ov, seconds):
            yield ov.phases.load_script
            time.sleep(seconds)
            yield ov.phases.run_script

        ...

        wait_five_seconds = wait_a_bit(seconds=5)

        ...

        ov.require(wait_five_seconds)  # <== will be called here

    Arguments:
        fn: The parametrized function. Its first argument, ov, will be given after
            the other arguments.
    """

    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        def instrument(ov):
            yield from fn(ov, *args, **kwargs)

        return instrument

    return wrapped


def configurable(fn):
    """Create a configurable instrument.

    The instrument must be a function with two parameters: the first one is the
    Overload and the second one must have a type annotation for a dataclass.
    The members of that dataclass will be added to the command-line arguments
    as per :meth:`~voir.argparse_ext.ExtendedArgumentParser.add_from_model`.
    """
    argspec = inspect.getfullargspec(fn)
    argname = argspec.args[1]
    ann = argspec.annotations[argname]

    @functools.wraps(fn)
    def wrapped(ov):
        yield ov.phases.init

        ov.argparser.add_from_model(argname, ann)

        yield ov.phases.parse_args
        yield from fn(ov, getattr(ov.options, argname))

    return wrapped



"""Utilities to split a script in a prelude and main part.

See :func:`split_script` for a more thorough description of what we are trying
to do here.
"""

import ast
import io
import sys
from importlib.machinery import ModuleSpec
from types import ModuleType


def resolve_script(script, module_name=None):
    """Return a function that calculates the main body of the script.

    Imports and functions are exec()ed by ``resolve_script``. Only the main body is
    not executed. The separation of the script in two parts is done via
    :func:`split_script`.

    Arguments:
        script: Path to the script.
        module_name: (optional) The name of the module that the script is supposed to
            represent.

    Returns:
        A nullary function that executes the script.
    """
    prep, mainsection = split_script(script)
    mod = ModuleType("__main__")
    glb = vars(mod)
    glb["__file__"] = script
    if module_name:
        glb["__spec__"] = ModuleSpec(name=module_name, loader=None)
    sys.modules["__main__"] = mod
    exec(prep, glb, glb)
    return lambda: exec(mainsection, glb, glb)


def split_script(script):
    """Split code that comes after all function definitions.

    Essentially, we want to be able to instrument functions in the main script, which
    requires evaluating the functions, but we want to do this before executing the main
    code. So we split off code that comes after function definitions so that we can evaluate
    the module and then evaluate that code separately.

    Code between function definitions will be evaluated right away, but the bulk usually
    comes after these definitions (because they need to use them).

    Arguments:
        script: Path to the script.

    Returns:
        A ``(prepare, run)`` tuple such that:

        * ``prepare()`` runs the import statements and function declarations up to the
          first statement.
        * ``run()`` runs the rest of the program after that.
    """

    with io.open_code(script) as f:
        source_code = f.read()

    tree = ast.parse(source_code, mode="exec")

    last_def = 0
    for i, stmt in enumerate(tree.body):
        if isinstance(stmt, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
            last_def = i + 1

    mod_before = ast.copy_location(
        ast.Module(
            body=tree.body[:last_def],
            type_ignores=[],
        ),
        tree,
    )

    mod_after = ast.copy_location(
        ast.Module(
            body=tree.body[last_def:],
            type_ignores=[],
        ),
        tree,
    )

    return (
        compile(mod_before, script, "exec"),
        compile(mod_after, script, "exec"),
    )



"""Monitor Network usage."""

import psutil


def network_monitor(pernic=False):
    def monitor():
        iocounters = psutil.net_io_counters(pernic=pernic)

        def netinfo(netio):
            return {
                "bytes_sent": netio.bytes_sent,
                "bytes_recv": netio.bytes_recv,
                "packets_sent": netio.packets_sent,
                "packets_recv": netio.packets_recv,
                "errin": netio.errin,
                "errout": netio.errout,
                "dropin": netio.dropin,
                "dropout": netio.dropout,
            }

        if pernic:
            return {str(k): netinfo(netio) for k, netio in iocounters.items()}

        return netinfo(iocounters)

    return monitor



import argparse


def collatz(n):
    while n != 1:
        print(n)
        n = (3 * n + 1) if n % 2 else (n // 2)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-n")
    options = parser.parse_args()
    collatz(int(options.n))


if __name__ == "__main__":
    main()



"""Command-line interface."""

import operator
import os
import sys
from functools import reduce
from pathlib import Path
from runpy import run_path

from ovld import ovld

from .overseer import Overseer

module = type(operator)


def find_voirfiles(script_path):
    """Find voirfiles that should be active for a given script.

    This proceeds by looking for ``voirfile.py`` in the directory and its
    parents.

    Arguments:
        script_path: The script's path.
    """
    script_path = Path(script_path).expanduser().absolute()
    cur = script_path

    if cur.is_file():  # pragma: no cover
        # Currently not used
        cur = cur.parent
        vf = (cur / script_path.stem).with_suffix(".voirfile.py")
        results = [vf]
    else:
        results = []

    while cur != cur.parent:
        results.append(cur / "voirfile.py")
        cur = cur.parent

    return [str(pth) for pth in results if pth.exists()]


@ovld
def _to_instruments(self, value: list):  # noqa: F811
    """Recursively find instruments in a data structure."""
    return reduce(operator.add, map(self, value), [])


@ovld
def _to_instruments(self, value: dict):  # noqa: F811
    return reduce(operator.add, map(self, value.values()), [])


@ovld
def _to_instruments(self, value):  # noqa: F811
    return [value]


def _collect_instruments(voirfile, i):
    """Collect instruments from a single voirfile.

    Arguments:
        voirfile: The path to the voirfile.
        i: An integer that determines the module name under which we will
           load the voirfile (``__voir0__``, ``__voir1__``, etc. because
           we want them to all coexist while avoiding clashes).
    """
    name = "__voir__" if i == 0 else f"__voir{i}__"
    md = module(name)
    md.__file__ = voirfile
    glb = run_path(voirfile, init_globals=vars(md), run_name=name)
    sys.modules[name] = md
    pfx = "instrument_"
    if "__instruments__" in glb:
        return _to_instruments(glb["__instruments__"])
    else:
        results = []
        for name, value in glb.items():
            if name.startswith(pfx):
                results.extend(_to_instruments(value))
        return results


def collect_instruments(voirfiles):
    """Collect instruments from a list of voirfiles.

    Arguments:
        voirfiles: List of files in which to find instruments.
    """
    return reduce(
        operator.add,
        [_collect_instruments(vf, i) for i, vf in enumerate(voirfiles)],
        [],
    )


def collect_contrib_instruments():
    """Collect instruments declared as entry points.

    This isn't really used, currently.
    """
    try:
        import pkg_resources

        results = []
        for entry_point in pkg_resources.iter_entry_points("voir.instrument"):
            results.append(entry_point.load())
        return results
    except ImportError:
        return []


def main(argv=None):
    """Entry point of the voir command line interface."""
    sys.path.insert(0, os.path.abspath(os.curdir))

    vfs = os.environ.get("VOIRFILE", None)
    if vfs is None:
        vfs = find_voirfiles(".")
    else:
        vfs = vfs.split()

    instruments = collect_instruments(vfs)
    instruments.extend(collect_contrib_instruments())

    ov = Overseer(instruments=instruments, logfile=int(os.environ.get("DATA_FD", 3)))
    ov(sys.argv[1:] if argv is None else argv)



import pytest


@pytest.mark.parametrize("prelude", (["voir"], ["python", "-m", "voir"]))
def test_cli(prelude, run_program):
    run_program([*prelude, "hello.py"])


def test_cli_dash_m(run_program):
    run_program(["voir", "-m", "hello:alt"])


def test_cli_package(run_program):
    run_program(["voir", "-m", "packpack"])


def test_environ(run_program):
    run_program(["voir", "hello.py"], voirfile="voirfile_nested.py")


def test_dunder_instruments(run_program):
    run_program(["voir", "hello.py"], voirfile="voirfile_dunder.py")


def test_forward(run_program):
    run_program(["voir", "giver.py"], voirfile="voirfile_fw.py")


def test_bad_unicode(run_program):
    run_program(["voir", "evil.py"])



import os
import subprocess

from .common import NotAvailable

IMPORT_ERROR = None
try:
    # python -m pip install --index-url https://pypi.anaconda.org/intel/simple dpctl
    import dpctl
except ImportError as err:
    IMPORT_ERROR = err


def query_gpu_data(gpu):
    # 0. GPU Utilization (%), GPU active time of the elapsed time, per tile or device. Device-level is the average value of tiles for multi-tiles.
    # 1. GPU Power (W), per tile or device.
    # 2. GPU Frequency (MHz), per tile or device. Device-level is the average value of tiles for multi-tiles.
    # 3. GPU Core Temperature (Celsius Degree), per tile or device. Device-level is the average value of tiles for multi-tiles.
    # 4. GPU Memory Temperature (Celsius Degree), per tile or device. Device-level is the average value of tiles for multi-tiles.
    # 5. GPU Memory Utilization (%), per tile or device. Device-level is the average value of tiles for multi-tiles.
    # 6. GPU Memory Read (kB/s), per tile or device. Device-level is the sum value of tiles for multi-tiles.
    # 7. GPU Memory Write (kB/s), per tile or device. Device-level is the sum value of tiles for multi-tiles.
    # 18. GPU Memory Used (MiB), per tile or device. Device-level is the sum value of tiles for multi-tiles.

    # xpu-smi does not seem to be working as expected
    output = subprocess.check_output(
        [
            "xpumcli",
            "dump",
            "-t",
            "0,1",  # All tiles, 1550 have 2 tiles
            "-d",
            "-1",  # All Devices
            "-m",
            "0,1,3,5,18",  # Compute Util, Power, Temp, Mem Util
            "-n",
            "1",  # Run once
        ],
        text=True,
    )

    def parse(val, type, default):
        try:
            return type(val)
        except ValueError:
            return default

    data = []
    total_size = gpu.max_mem_alloc_size / (1024 * 1024)

    for i, line in enumerate(output.split("\n")):
        if i == 0:
            continue

        if len(line) == 0:
            continue

        (
            timestamp,
            device_id,
            tile_id,
            gpu_util,
            power,
            temp,
            mem_per,
            mem_bytes,
        ) = line.split(",")
        device_id = parse(device_id, int, 0)
        tile_id = parse(tile_id, int, 0)

        data.append(
            {
                "device": f"level_zero:{device_id * 2 + tile_id}",
                "product": gpu.name,
                "memory": {
                    "used": parse(mem_bytes, float, 0),
                    "total": total_size,
                },
                "utilization": {
                    "compute": parse(gpu_util, float, 0) / 100,
                    "memory": parse(mem_per, float, 0) / 100,
                },
                "temperature": parse(temp, float, 0),
                "power": parse(power, float, 0),
                "selection_variable": "ONEAPI_DEVICE_SELECTOR",
            }
        )
    return data


def get_devices():
    return dpctl.get_devices()


def get_gpus():
    gpus = []
    cpus = []

    for device in get_devices():
        # GPUs are shown as level_zero AND openCL
        if device.is_gpu and "level_zero" in device.get_filter_string():
            gpus.append(device)

        if device.is_cpu:
            cpus.append(device)

    return gpus


def is_installed():
    return IMPORT_ERROR is None


class DeviceSMI:
    def __init__(self) -> None:
        if IMPORT_ERROR is not None:
            raise NotAvailable from IMPORT_ERROR

        self.gpus = get_gpus()
        visible_devices = os.environ.get("ONEAPI_DEVICE_SELECTOR", None)

        if visible_devices is None:
            self.device_ids = list(range(len(self.gpus)))
        else:
            device_ids = []
            for device in visible_devices.split(","):
                device_ids.append(int(device.split(":")[-1]))

            self.device_ids = device_ids

    @property
    def arch(self):
        return "xpu"

    @property
    def visible_devices(self):
        return os.environ.get("ONEAPI_DEVICE_SELECTOR", None)

    def get_gpus_info(self, selection=None):
        # Assume all GPUs are the same
        all_gpus = query_gpu_data(self.gpus[0])

        return {i: all_gpus[i] for i in self.device_ids}

    def close(self):
        pass

    def system_info(self):
        try:
            # untested
            driver_version = "NA"
            gfx_firmware_version = "NA"
            gfx_data_firmware_version = "NA"

            output = subprocess.check_output(
                [
                    "xpumcli",
                    "dump",
                    "-t",
                    "0,1",  # All tiles, 1550 have 2 tiles
                    "-d",
                    "-1",  # All Devices
                    "-m",
                    "8,9,10",
                    "-n",
                    "1",  # Run once
                ],
                text=True,
            )
            for i, line in enumerate(output.split("\n")):
                if i == 0:
                    continue

                if len(line) == 0:
                    continue

                (
                    driver_version,
                    gfx_firmware_version,
                    gfx_data_firmware_version,
                ) = line.split(",")

            return {
                "DRIVER_VERSION": driver_version,
                "GFX_FIRMWARE_VERSION": gfx_firmware_version,
                "GFX_DATA_VERSION": gfx_data_firmware_version,
            }
        except Exception:
            import traceback

            traceback.print_exc()
            return {}



"""Monitor CPU usage."""

from collections import defaultdict

import psutil


def cpu_monitor():
    def monitor():
        mem = psutil.virtual_memory()
        return {
            "memory": [
                mem.used,
                mem.total,
            ],
            "load": psutil.cpu_percent(),
        }

    return monitor


def retrieve_process_info(process, acc):
    try:
        io = process.io_counters()
        acc["read_bytes"] += io.read_bytes
        acc["write_bytes"] += io.write_bytes
        acc["read_chars"] += io.read_chars
        acc["write_chars"] += io.write_chars

        mem = process.memory_info()
        acc["mem_used"] += mem.rss

        cpu = process.cpu_percent(interval=1)
        acc["cpu_percent"] += cpu
        acc["children"] += 1
    except psutil.AccessDenied:
        pass
    except psutil.NoSuchProcess:
        pass


def _recursive(process, acc):
    for child in process.children(recursive=True):
        retrieve_process_info(child, acc)


def process_monitor(pid, recursive=True):
    process = psutil.Process(pid)
    cpu_count = psutil.cpu_count()
    mem_total = psutil.virtual_memory()

    def monitor():
        nonlocal process
        acc = defaultdict(float)

        with process.oneshot():
            cpu_num = process.cpu_num()
            retrieve_process_info(process, acc)
            if recursive:
                _recursive(process, acc)

        return {
            "pid": pid,
            # CPU
            "load": acc["cpu_percent"] / cpu_count,
            "num": cpu_num,
            # IO
            "read_bytes": acc["read_bytes"],
            "write_bytes": acc["write_bytes"],
            # Fake IO
            "read_chars": acc["read_chars"],
            "write_chars": acc["write_chars"],
            # Memory
            "memory": [acc["mem_used"], mem_total.total],
        }

    return monitor



"""Utilities to smuggle data into stdout.

The basic idea is to print out a (syntactically valid but meaningless) escape
sequence that terminals will ignore, that way the output looks fine on the
terminal, but there is extra data embedded in the stream that can be recovered.
"""

import re
from collections import deque

# Standard hexadecimal characters
oldhex = "0123456789abcdef"

# Arbitrary sequences of these sixteen characters are valid in escape sequences
newhex = "0123456789:;<=>?"

hexsub = dict(zip(oldhex, newhex))
hexsub_back = dict(zip(newhex, oldhex))


def encode_as_escape_sequence(s: str):
    """Encode a string as a CSI escape sequence.

    The escape sequence is meaningless, but terminals will ignore it, so we can
    smuggle data invisibly that way.

    A CSI code starts with the ESC character followed by [, an arbitrary number
    of characters in the 0x30 to 0x3F range (see the newhex variable), then 0 or more
    characters in the 0x20 to 0x2F range and then one byte from 0x40-0x7E.

    We simply encode the data as hex using the characters in 0x30-0x3F, to which we
    append the letter z to end the escape sequence.

    Arguments:
        s: The string to encode.
    """
    encoded = "".join([hexsub[c] for c in s.encode("utf8").hex()])
    return f"\033[{encoded}z"


def decode_escape_sequence(s):
    """Decode an escape sequence generated by :func:`encode_as_escape_sequence`."""
    s = s[2:-1]
    decoded_hex = "".join([hexsub_back[c] for c in s])
    return bytes.fromhex(decoded_hex).decode("utf8")


class SmuggleWriter:
    """Class for smuggling data into a stream using escape sequences.

    This uses :func:`encode_as_escape_sequence` on each write.

    Arguments:
        stream: The stream in which to smuggle the data, e.g. ``sys.stdout``
    """

    def __init__(self, stream):
        self.stream = stream

    def __enter__(self):
        self.stream.__enter__()
        return self

    def __exit__(self, typ=None, exc=None, tb=None):
        self.stream.__exit__(typ, exc, tb)
        return self

    def write(self, txt):
        return self.stream.write(encode_as_escape_sequence(txt))

    def flush(self):
        return self.stream.flush()

    def close(self):
        return self.stream.close()


class LineAccumulator:
    """Process individual characters and accumulate full lines into a deque."""

    def __init__(self):
        self.lines = deque()
        self.current = ""

    def process(self, char):
        self.current += char
        if char == "\n":
            self.lines.append(self.current)
            self.current = ""


def utf8_length(char: chr) -> int:
    """Return number of bytes a UTF-8 character uses from its first byte."""
    b = ord(char)

    if b & 0b10000000 == 0:
        return 1  # ASCII
    elif b & 0b11100000 == 0b11000000:
        return 2
    elif b & 0b11110000 == 0b11100000:
        return 3
    elif b & 0b11111000 == 0b11110000:
        return 4
    return 1


class Decoder:
    """Decoder for a stream where data has been smuggled.

    Produces two :class:`LineAccumulator`s, ``out`` and ``data``:

    * ``out`` contains the normal output of the stream.
    * ``data`` contains everything that was encoded through
      :func:`encode_as_escape_sequence`

    Arguments:
        principal: The stream.
    """

    def __init__(self, principal):
        self.principal = principal
        self.out = LineAccumulator()
        self.data = LineAccumulator()
        self.code = ""
        self.coding = False

    def close(self):
        self.principal.close()

    def endcode(self):
        self.coding = False
        if re.match(string=self.code, pattern="\033\\[[0-9:;<=>?]*z"):
            for char in decode_escape_sequence(self.code):
                self.data.process(char)
        else:
            self.out.current += self.code
        self.code = ""

    def process_char(self, char):
        if self.coding:
            if ord(char) < 0x20:
                self.out.process(char)
                self.endcode()
            elif ord(char) >= 0x40 and char != "[":
                self.code += char
                self.endcode()
            else:
                self.code += char

        elif char == "\033":
            self.coding = True
            self.code += char

        else:
            self.out.process(char)

    def getline(self, which):
        if which == "out":
            if self.out.lines:
                return self.out.lines.popleft()
        elif which == "data":
            if self.data.lines:
                return self.data.lines.popleft()
        return None

    def readline(self, which: str):
        """Get the next line from either ``self.out`` or ``self.data``.

        Arguments:
            which: Either ``"out"`` or ``"data"``.

        Returns:
            The next line or ``None`` if there is no data.
        """
        while (result := self.getline(which)) is None:
            nxt = self.principal.read(1)
            if not nxt:
                return None

            for i in range(utf8_length(nxt) - 1):
                nxt += self.principal.read(1)

            self.process_char(nxt.decode("utf8"))

        return result


class MultimodalFile:
    """File object for a stream in which data has been smuggled.

    Arguments:
        decoder: The :class:`Decoder` instance to read from.
        which: Either ``"out"`` or ``"data"``.
    """

    def __init__(self, decoder: Decoder, which: str, name: str):
        self.name = name
        self.decoder = decoder
        self.which = which

    def close(self):
        self.decoder.close()

    def fileno(self):
        return self.name

    def readline(self):
        return self.decoder.readline(self.which)



import json
import os
import sys

special_fd = int(os.environ.get("DATA_FD", 3))

if __name__ == "__main__":
    print("to stdout")
    print("to stderr", file=sys.stderr)
    with open(special_fd, "w", buffering=1) as lg:
        for i in range(5):
            print(json.dumps({"message": i}), file=lg)
        print("gargle gargle", file=lg)
    print("to stdout again")



"""Monitor GPU usage."""

import time

from ..tools import instrument_definition
from .cpu import cpu_monitor
from .gpu import gpu_monitor as gpu_monitor_fun, select_backend
from .io import io_monitor
from .network import network_monitor
from .utils import monitor as generic_monitor


def monitor(ov, poll_interval=10, worker_init=None, **monitors):
    """Monitor metrics given by monitors"""

    yield ov.phases.load_script

    def get():
        t = time.time()
        entries = []
        for k, v in monitors.items():
            values = {
                "task": "main",
                "time": t,
                k: v(),
            }
            entries.append(values)
        return entries

    def push(data):
        for entry in data:
            ov.give(**entry)

    mon = generic_monitor(
        poll_interval,
        get,
        push,
        process=False,
        worker_init=worker_init,
    )
    mon.start()
    try:
        yield ov.phases.run_script
    finally:
        mon.stop()


@instrument_definition
def monitor_all(ov, poll_interval=10, arch=None):
    return monitor(
        ov,
        poll_interval=poll_interval,
        gpudata=gpu_monitor_fun(),
        iodata=io_monitor(),
        netdata=network_monitor(),
        cpudata=cpu_monitor(),
        worker_init=lambda: select_backend(arch, force=True),
    )


@instrument_definition
def gpu_monitor(ov, poll_interval=10, arch=None):
    """Monitor GPU utilization.

    Supports monitoring CUDA (NVIDIA) and ROCm (AMD) architectures.

    The following data is monitored:

    .. code-block:: javascript

        {
            "memory": [USED, TOTAL],  // In MB
            "load": LOAD,             // Utilization, from 0 to 1
            "temperature": TEMP,      // In celsius
            "power": POWER,
        }

    This data structure is added to the :meth:`~voir.overseer.Overseer.given` stream
    as follows:

    .. code-block:: python

        give(task="main", gpudata=DATA)

    Arguments:
        poll_interval: The polling interval, in seconds. Data will be produced
            every poll_interval seconds.
        arch: The GPU architecture to monitor. If None, the architecture will be
            deduced automatically.
    """
    return monitor(
        ov,
        poll_interval=poll_interval,
        gpudata=gpu_monitor_fun(),
        worker_init=lambda: select_backend(arch, force=True),
    )
